/****************************************************************************/
/* Copyright 2002 Compaq Computer Corporation.                              */
/*                                           .                              */
/* Copying or modifying this code for any purpose is permitted,             */
/* provided that this copyright notice is preserved in its entirety         */
/* in all copies or modifications.  COMPAQ COMPUTER CORPORATION             */
/* MAKES NO WARRANTIES, EXPRESSED OR IMPLIED, AS TO THE USEFULNESS          */
/* OR CORRECTNESS OF THIS CODE OR ITS FITNESS FOR ANY PARTICULAR            */
/* PURPOSE.                                                                 */
/****************************************************************************/
/*
 * IDE drive support
 */

#include <errno.h>
#include "bootldr.h"
#include "heap.h"
#include "fs.h"
#include "ide.h"
#include "disk-partition-table.h"

enum ide_registers {
   IDE_DATA_REG    = 0x0,
   IDE_WRITE_PRECOMPENSATION_REG = 0x1,
   IDE_SECTOR_COUNT_REG = 0x2,
   IDE_SECTOR_NUMBER_REG = 0x3,
   IDE_CYLINDER_LOW_REG = 0x4,
   IDE_CYLINDER_HIGH_REG = 0x5,
   IDE_DRIVE_HEAD_REG = 0x6,
   IDE_COMMAND_REG = 0x7,
   IDE_STATUS_REG  = 0x7,
   IDE_SECTOR_BUF  = 0x400
};

enum ide_command {
   IDE_COMMAND_READ_BUFFER  = 0xE4,
   IDE_COMMAND_READ_SECTORS = 0x20,
   IDE_COMMAND_IDENTIFY = 0xEC,
};

/* set this bit if passing LBA instead of C/H/S */
#define DEVICE_HEAD_IS_LBA (1 << 6)

struct drive_identification {
  /* 0x00 */ unsigned short flags;
  /* 0x01 */ unsigned short n_cylinders;
  /* 0x02 */ unsigned short reserved02;
  /* 0x03 */ unsigned short n_heads;
  /* 0x04 */ unsigned short n_bytes_per_track;
  /* 0x05 */ unsigned short n_bytes_per_sector;
  /* 0x06 */ unsigned short n_sectors_per_track;
  /* 0x07 */ unsigned short inter_sector_gap;
  /* 0x08 */ unsigned short reserved08;
  /* 0x09 */ unsigned short n_vendor_status_words;
  /* 0x0a */ unsigned char serial_number[20];
  /* 0x14 */ unsigned short controller_type;
  /* 0x15 */ unsigned short n_buffer_512b;
  /* 0x16 */ unsigned short n_ecc_bytes;
  /* 0x17 */ unsigned char firmware_revision[8];
  /* 0x1b */ unsigned char model_number[40] ;
  /* 0x2f */ unsigned short read_write_multiples_implemented;
  /* 0x30 */ unsigned short double_word_io_supported;
  /* 0x31 */ unsigned short capabilities31;
  /* 0x32 */ unsigned short capabilities32;
  /* 0x33 */ unsigned short minimum_pio_mode_number;
  /* 0x34 */ unsigned short retired34;
  /* 0x35 */ unsigned short flags53;
  /* 0x36 */ unsigned short current_n_cylinders;
  /* 0x37 */ unsigned short current_n_heads;
  /* 0x38 */ unsigned short current_n_sectors_per_track;
  /* 0x39 */ unsigned long  current_n_sectors;
  /* 0x3a */
  /* 0x3b */ unsigned short reserved3b;
  /* 0x3c */ unsigned long n_sectors_user_addressable;
  /* 0x3e */
};

struct ide_adapter {
   volatile char *ioport;
   char buf[512];
   int ptable_was_read;
   struct dos_ptable_entry ptable[16];
   struct drive_identification identification; 
};

static struct ide_adapter *ide_adapter = 0;

static __inline unsigned short short_byte_swap(unsigned short x)
{
  return ((x & 0xFF) << 8) | ((x & 0xFF00) >> 8);
}
 
int ide_identify(void)
{
   int i;
   if (ide_adapter) {
      char serial_number[21];
      char firmware_revision[9];
      char model_number[40];
      ide_adapter->ioport[IDE_COMMAND_REG] = IDE_COMMAND_IDENTIFY;
      while (ide_adapter->ioport[IDE_STATUS_REG] & 0x80) { 
         putLabeledWord(" ide status=", ide_adapter->ioport[IDE_STATUS_REG]);
      }
      /* copy in identification */
      for (i = 0; i < sizeof(struct drive_identification)/2; i++) { 
         unsigned short *buf = (unsigned short *)&ide_adapter->identification;
         unsigned short *sector_buffer = (unsigned short *)&ide_adapter->ioport[IDE_SECTOR_BUF];
         buf[i] = sector_buffer[i];
      }
      hex_dump((unsigned char *)&ide_adapter->identification, sizeof(struct drive_identification));
      memset(serial_number, 0, sizeof(serial_number)); 
      memcpy(serial_number, ide_adapter->identification.serial_number, sizeof(serial_number)-1);
      memset(firmware_revision, 0, sizeof(firmware_revision));
      memcpy(firmware_revision, ide_adapter->identification.firmware_revision, sizeof(firmware_revision)-1);
      memset(model_number, 0, sizeof(model_number));
      memcpy(model_number, ide_adapter->identification.model_number, sizeof(model_number)-1);
      putstr("  serial_number: "); putstr(serial_number); putstr("\r\n");
      putstr("  firmware_revision: "); putstr(firmware_revision); putstr("\r\n");
      putstr("  model_number: "); putstr(model_number); putstr("\r\n");
      putLabeledWord("  bytes_per_sector=", ide_adapter->identification.n_bytes_per_sector);
      putLabeledWord("  bytes_per_track=", ide_adapter->identification.n_bytes_per_track);
      putLabeledWord("  n_sectors_per_track=", ide_adapter->identification.n_sectors_per_track);
      putLabeledWord("  n_cylinders=", ide_adapter->identification.n_cylinders);
      putLabeledWord("  n_heads=", ide_adapter->identification.n_heads);
      putLabeledWord("  n_sectors_user_addressable=", ide_adapter->identification.n_sectors_user_addressable);
      return 0;
   } else {
     return -EINVAL;
   }
}

int ide_attach(volatile char *ioport)
{
  if (!ide_adapter)
    ide_adapter = mmalloc(sizeof(struct ide_adapter));
  if (!ide_adapter)
    return -ENOMEM;
  memset(ide_adapter, 0, sizeof(struct ide_adapter));
  ide_adapter->ioport = ioport;
  ide_identify();
  return 0;
}

int ide_detach(volatile char *ioport)
{

  return 0;
}

int ide_read(char *buf, unsigned long offset, unsigned long len)
{
  if (!ide_adapter || !ide_adapter->identification.n_bytes_per_sector) {
    return -EINVAL;
  } else {
    unsigned long sector_count = len / ide_adapter->identification.n_bytes_per_sector;
    unsigned long sector_number = offset / ide_adapter->identification.n_bytes_per_sector;
    unsigned long start_sector = sector_number & 0xff;
    unsigned long start_cylinder = (sector_number >> 8) &0xffff ;
    unsigned long start_head = (sector_number >> 24) & 0x7 ;
    int i;
#if 0
    putLabeledWord("  start_cylinder=", start_cylinder);
    putLabeledWord("  start_head=", start_head);
    putLabeledWord("  start_sector=", start_sector);
    putLabeledWord("  sector_count=", sector_count);
#endif
    ide_adapter->ioport[IDE_SECTOR_NUMBER_REG] = start_sector;
    ide_adapter->ioport[IDE_CYLINDER_HIGH_REG] = (start_cylinder >> 8) & 0xff;
    ide_adapter->ioport[IDE_CYLINDER_LOW_REG] = start_cylinder & 0xff;
    ide_adapter->ioport[IDE_SECTOR_COUNT_REG] = sector_count & 0xff;
    ide_adapter->ioport[IDE_DRIVE_HEAD_REG] = DEVICE_HEAD_IS_LBA | (start_head & 0xf);
    if (0) putstr("issuing read command\r\n");
    ide_adapter->ioport[IDE_COMMAND_REG] = IDE_COMMAND_READ_SECTORS;
    while (ide_adapter->ioport[IDE_STATUS_REG] & 0x80) { 
    }
    if (0) putLabeledWord(" ide status=", ide_adapter->ioport[IDE_STATUS_REG]);
    for (i = 0; i < 512; i++) { 
      buf[i] = ide_adapter->ioport[IDE_SECTOR_BUF + i];
    }
   
    return 0;
  }
}

int ide_read_ptable(struct dos_ptable_entry *ptable)
{
   unsigned char *buf = (unsigned char *)ide_adapter->buf;
   int result = 0;
   if (!ide_adapter)
      return -ENOMEM;
   result = ide_read(buf, 0, 0x200);
   if (result)
      return result;
   putLabeledWord("ptable signature=", (buf[511] << 8) | buf[510]);
   if (buf[510] != 0x55 || buf[511] != 0xaa) {
      putstr("did not find dos partition table signature\r\n");
      return -EINVAL;
   }
   memcpy(ide_adapter->ptable, &ide_adapter->buf[446], 4 * sizeof(struct dos_ptable_entry));
   ide_adapter->ptable_was_read = 1;
   if (ptable)
      memcpy(ptable, &ide_adapter->buf[446], 4 * sizeof(struct dos_ptable_entry));
   return 0;
}

/* returns start offset in bytes */
int ide_get_partition_start_byte(int partid)
{
   if (!ide_adapter)
      return -ENODEV;
   if (partid > 4)
      return -EINVAL;
   if (!ide_adapter->ptable_was_read)
      ide_read_ptable(0);
   return ide_adapter->ptable[partid].starting_sector_lba * ide_adapter->identification.n_bytes_per_sector;
}
int ide_get_partition_n_bytes(int partid)
{
   if (!ide_adapter)
      return -ENODEV;
   if (partid > 4)
      return -EINVAL;
   if (!ide_adapter->ptable_was_read)
      ide_read_ptable(0);
   return ide_adapter->ptable[partid].n_sectors * ide_adapter->identification.n_bytes_per_sector;
}


int ide_read_partition(char *buf, int partid, size_t *lenp)
{
  struct dos_ptable_entry *ptable = ide_adapter->ptable;
  unsigned long n_bytes_per_sector = 0;
  if (!ide_adapter->ptable_was_read) 
    ide_read_ptable(0);
  n_bytes_per_sector = ide_adapter->identification.n_bytes_per_sector;
  putstr("reading partition\r\n");
  putLabeledWord("buf=", (u32)buf);
  putLabeledWord("partid=", partid);
  putLabeledWord("  start_sector=", ptable[partid].starting_sector_lba);
  putLabeledWord("  n_sectors   =", ptable[partid].n_sectors);
  putLabeledWord("  n_bytes     =", ptable[partid].n_sectors * n_bytes_per_sector);
  if (buf) {
    unsigned long sector_size = n_bytes_per_sector;
    unsigned long sector;
    unsigned long start_sector = ptable[partid].starting_sector_lba;
    unsigned long n_sectors = ptable[partid].n_sectors;
    unsigned long end_sector = start_sector + n_sectors;
    unsigned long offset = 0;
    if (lenp) {
      if (*lenp && (*lenp < n_sectors * sector_size)) {
        n_sectors = *lenp / sector_size;
        end_sector = start_sector + n_sectors;
        putLabeledWord(" reading n_sectors ", n_sectors);
      }
      *lenp = n_sectors * sector_size;
    }
    for (sector = start_sector; sector < end_sector; sector++) {
      ide_read(buf+offset, sector*sector_size, sector_size);
#if 0 
      if ((offset & 0x3ffff) == 0) {
          putLabeledWord(" sector=", sector);
          putLabeledWord(" offset=", offset);
          putLabeledWord("  value=", *(long*)(buf+offset));
      }
#endif
      offset += sector_size;
    } 
  }      
  return 0;
}

static int ide_iohandle_read(struct iohandle *ioh, char *buf, size_t offset, size_t len) 
{
   struct dos_ptable_entry *pentry = (struct dos_ptable_entry *)ioh->pdata;
   unsigned long sector_size = ide_adapter->identification.n_bytes_per_sector;
   unsigned long start_sector = pentry->starting_sector_lba;
   return ide_read(buf, offset + start_sector * sector_size, len);
}

static int ide_iohandle_write(struct iohandle *ioh, char *buf, size_t offset, size_t len)
{
   return -EINVAL;
}

static int ide_iohandle_close(struct iohandle *ioh)
{
   return 0;
}

static struct iohandle_ops ide_iohandle_ops = {
   read: ide_iohandle_read,
   write: ide_iohandle_write,
   close: ide_iohandle_close
};

struct iohandle *get_ide_partition_iohandle(u8 partid)
{
  struct iohandle *ioh = mmalloc(sizeof(struct iohandle));
   if (!ide_adapter->ptable_was_read)
      ide_read_ptable(0);
  ioh->ops = &ide_iohandle_ops;
  ioh->pdata = &ide_adapter->ptable[partid];
  return ioh;
} 

void command_ide(int argc, const char* argv[])
{
   if (strcmp(argv[1], "identify") == 0) {
     ide_identify();
   } else if (strcmp(argv[1], "attach") == 0) {
      volatile char *ioport = (volatile char *)strtoul(argv[2], 0, 0);
      putLabeledWord("attaching ide drive at ioport=", (u32)ioport);
      ide_attach(ioport);
   } else if (strcmp(argv[1], "read") == 0) {
      char *buf = (char *)strtoul(argv[2], 0, 0);
      unsigned long start_byte = strtoul(argv[3], 0, 0);
      unsigned long n_bytes = strtoul(argv[4], 0, 0);
      ide_read(buf, start_byte, n_bytes);
   } else if (strcmp(argv[1], "ptable") == 0) {
      struct dos_ptable_entry ptable[4]; 
      if (!ide_read_ptable(ptable)) {
         hex_dump((unsigned char *)ptable, sizeof(ptable));
      }
   } else if (strcmp(argv[1], "read_sector") == 0) {
      char *buf = (char *)strtoul(argv[2], 0, 0);
      unsigned long sector = strtoul(argv[3], 0, 0);
      ide_read(buf, sector * 0x200, 0x200);
   } else if (strcmp(argv[1], "read_partition") == 0) {
      char *buf = (char *)strtoul(argv[2], 0, 0);
      int partid = strtoul(argv[3], 0, 0);
      size_t len = 0;
      if (argv[4] != NULL) {
        len = strtoul(argv[4], 0, 0);
        putLabeledWord("  len=", len);
      }
      ide_read_partition(buf, partid, &len);
   } else {
     putstr("usage:\r\n");
     putstr("    ide identify | attach <ioaddr>\r\n");
     putstr("    ide read <dstaddr> <byteoffset> <nbytes>\r\n");
     putstr("    ide read_sector <dstaddr> <sectorno>\r\n");
     putstr("    ide read_partition <dstaddr> <partitionno> [<nbytes>]\r\n");
   }
}
